package structures

import (
	"ORsearch/consts"
)



// *** COMPATIBLE PAIRS RELATED FUNCTIONS ***

// compatiblepairs[l][r][i] is the i^th possible q value that for a valid pair with r 
//     such that (r,q) is compatible with the rows 0 to l of the matrix m
// cardinalityQr[l][r] is the number of entries in the Q set compatiblepairs[l][r]



// Initializes the structures compatiblepairs and cardinalityQr
// and populate them according to the already filled lines of the matrix m
func InitCompatiblePairs(m *Matrix, fixedrows int) ([][][]uint16, [][]int, bool) {

	// Initialize the big m x 2^n table : the set of compatible pairs P for each matix level
	compatiblepairs := make([][][]uint16, ConfigureM(consts.N))
	for l := range compatiblepairs {
		compatiblepairs[l] = make([][]uint16, 1<<uint(consts.N))
		for r := range compatiblepairs[l] {
			compatiblepairs[l][r] = make([]uint16, 1<<uint(consts.N))
		}
	}
	
	// Initialize the cardinalities of all Q(r) sets
	cardinalityQr := make([][]int, ConfigureM(consts.N))
	for l := range cardinalityQr {
		cardinalityQr[l] = make([]int, 1<<uint(consts.N))
	}

	// Initialize the first level to the full compatible set
	for r := range compatiblepairs[0] {
		for i := range compatiblepairs[0][r] {
			compatiblepairs[0][r][i] = uint16(i)
		}
		cardinalityQr[0][r] = 1<<uint(consts.N)
	}

	// Make fixedrows-1 calls to UpdateCompatiblePairs
	goodbranch := true
	for l := 1; l < fixedrows; l++ {
		if goodbranch {
			goodbranch = UpdateCompatiblePairs(compatiblepairs, cardinalityQr, m, l)
		}
	}

	return compatiblepairs, cardinalityQr, goodbranch
	
}



// Now that we added a new row at level limitlevel, we compute the pairs of rows that are compatible with the matrix m
// These pairs should be a subset of the previous set of campatible pairs (stored in compatiblepairs[limitlevel-1])
// Return true if the cutting lemma is satisfied
func UpdateCompatiblePairs(compatiblepairs [][][]uint16, cardinalityQr [][]int, m *Matrix, limitlevel int) bool {

	var prev01 uint16 = (^m.Elems[limitlevel-1]) & (m.Elems[limitlevel])
	var prev10 uint16 = (m.Elems[limitlevel-1]) & (^m.Elems[limitlevel])

	// If the two last rows are equal, no use to go further
	if prev01 == 0 && prev10 == 0 {
		return false
	}

	var r, q uint16
	var cardinalityR, i int

	// We browse all the previously possible pairs (r,q), that is, the ones in compatiblepairs[limitlevel-1]
	for r = 0; r < 1<<uint(consts.N); r++ {
		cardinalityQr[limitlevel][r] = 0
		// First check : is there an r that works? If so, then there will also be a q (since we can take q = r)
		if ((prev01 & r) | (prev10 & (^r)) != 0) && (cardinalityQr[limitlevel-1][r] > 0) {
			cardinalityR++
			// We browse the possible q-values from the previous compatiblepairs (since the new ones must be a subset)
			for i = 0; i < cardinalityQr[limitlevel-1][r]; i++ {
				q = compatiblepairs[limitlevel-1][r][i]
				// If the q-value from the previous compatiblepairs works, then we add it to the new Q set
				if (prev01 & r & q) | (prev10 & (^r) & (^q)) != 0 {
					compatiblepairs[limitlevel][r][cardinalityQr[limitlevel][r]] = q
					cardinalityQr[limitlevel][r]++
				}
			}
		}
	}

	return (limitlevel + cardinalityR >= consts.CUTATUPDATE)

}