package structures

import (
	"ORsearch/consts"
	"ORsearch/utilities"
)



// *** SYMMETRY RELATED FUNCTIONS ***

// A matrix m allows for symmetry if some of its columns are identical.
// Indeed, extensions of m can provide identical matrices up to column permutations.
// To avoid that, we impose that the columns are sorted lexicographically
// (where the left most column (that corresponds to the largest bit of the rows) has the smallest "value")
// That way, identical columns are always next to each other. 
// For any matrix m, we define the symmetry vector sym(m) that allows to track back the identical columns:
//		sym(m)_N = 1 (always)
// 		if sym(m)_i = 0, then m_{:,i} = m_{:,i+1}
// Therefore, m does not allow for symmetry iff sym(m) = 11...1.



// Given a matrix m, returns the symmetry vector for each submatrix 
// m^i = [m_{0,:} ; ... ; m_{i,:}], 0 <= i < fixedrows
func InitSymmetry(m *Matrix, fixedrows int) []uint16 {
	// sym[i] corresponds to m^i
	var sym = make([]uint16, ConfigureM(consts.N))
	sym[0] = 1
	for i := 1 ; i < fixedrows ; i++ {
		sym[i] = UpdateSymmetry(sym[i-1], m.Elems[i])
	}
	return sym
}



// Suppose sym is the symmetry vector corresponding to a matrix m and we would like to add a row newrow to m.
// Then UpdateSymmetry(sym, newrow) returns the symmetry vector of the resulting matrix.
func UpdateSymmetry(sym, newrow uint16) uint16 {
	if IsSym(sym) {
		for i := 1 ; i < consts.N ; i++ {
			if !utilities.GetBit(sym, uint(i)) && !utilities.CompareBits(newrow, uint(i), uint(i-1)) {
				sym += 1 << uint(i)
			}
		}
	} 
	return sym
}



// Does the symmetry vector allow for symmetry?
// True if sym != 11...1
func IsSym(sym uint16) bool {
	return !(sym == 1<<uint(consts.N)-1)
}



// Let sym be the symmetry vector corresponding to some matrix m.
// Returns true if the matrix [m ; row] satisfies the symmetry rules
func IsSymOK(sym, row uint16) bool {
	var maxbit = true
	for i := 0; i < consts.N; i++ {
		if utilities.GetBit(sym, uint(i)) {
			maxbit = true
		}
		if !utilities.GetBit(row, uint(i)) {
			maxbit = false
		}
		if utilities.GetBit(row, uint(i)) && !maxbit {
			return false
		}
	}
	return true
}