package workers

import (
	"fmt"
	"math/rand"
	"time"
	"ORsearch/consts"
	"ORsearch/structures"
	"ORsearch/parsedata"
)



// Compute each branch of length consts.DISTRIBUTIONLEVEL on which to launch the branch search algorithm
// The jobs are attributed to workers in a random order from randomLaunchWorkers()
//
// consts.INITIALROWS defines the number of rows that are fixed for all branches according to structures.InitialRows()
// (The first and second rows can be set to all zeros and all ones respectively WLOG)
// (For all n <= 7, the 4 first rows of the extremal matrices always have the same structure)
//
// Workers are only launched on the relevant branches, that is:
// - those where the matrix columns are sorted lexicographically (no symmetry)
// - those that should not yet be cut (see the cutting lemma)
// - those that were not yet fully explored in a previous (interrupted) run of the program 
//   (those that are not in branchesdone)
func StartWorkers(freethread chan structures.Empty, res chan structures.SearchResult, branchesdone [][]uint16) {

	// *** INITIALIZATION ***

	// Initialize the matrix with the chosen initial rows
	var Init = structures.InitialRows(consts.INITIALROWS)
	var m = structures.InitializeMatrix(Init)

	// Initialize the compatible pairs and the symmetry for these rows
	var compatiblepairs, cardinalityQr, _ = structures.InitCompatiblePairs(m, consts.INITIALROWS-1)
	// We only initialize for the consts.INITIALROWS-1 first rows, the last one being settled by branchsearch()
	var sym = structures.InitSymmetry(m, consts.INITIALROWS)

	// Initialize the list of branches
	var branchlist = make([][]uint16, 0)

	// We have to initialize these even though they are useless here
	var bestlength = consts.DISTRIBUTIONROWS-1
	var bestmatrix = structures.NewMatrix(0,0)

	// *** LAUNCH THE SEARCH ***

	// Report every branch of size consts.DISTRIBUTIONROWS in branchlist, according for symmetry and cutting
	branchSearch(consts.INITIALROWS, m, compatiblepairs, cardinalityQr, &bestlength, bestmatrix, consts.DISTRIBUTIONROWS, &branchlist, sym)

	// *** FILTERING OF THE OLD BRANCHES ***

	// Filter out branchlist and remove all entries that have already been explored
	var workerno = 0
	var filteredbranchlist = make([][]uint16, 0, len(branchlist))
	for i := range branchlist {
		if !parsedata.IsInBranchlist(branchlist[i], branchesdone) {
			filteredbranchlist = append(filteredbranchlist, branchlist[i])
			workerno++
		}
	}

	if consts.DISPLAY >= 0 {
		fmt.Println("Number of branches:", len(filteredbranchlist))
	}

	// *** SEND THE WORKERS ***

	randomLaunchWorkers(filteredbranchlist, freethread, res, workerno)

}



// For each relevant branch of size consts.DISTRIBUTIONROWS computed by Startworkers(), launch a worker in random 
// order
//
// Exploring the branches at random allows to have a better idea of the overall progress
// by looking at the ratio between the number of branches done and the total number of branches to explore.
// Otherwise, there is no guarantee that the short/long branches are not searched first for instance.
func randomLaunchWorkers(brancheslist [][]uint16, freethread chan structures.Empty, res chan structures.SearchResult, workerno int) {

	rand.Seed(time.Now().UnixNano())

	// Start the first few branches, nonblocking!
	go func() { for j := 0; j < consts.PARALLELTHREADS; j++ { freethread <- structures.Empty{} } }()

	// Keep the number of running threads constant
	for i := 0; i < workerno; i++ {
		<- freethread

		//Display progress if requested
		if consts.DISPLAY >= 1 && i % consts.PROGRESSFREQUENCY == 0 {
			fmt.Println("Progress: ",i,"/",workerno, "(",float64(i*1000/workerno)/10,"%)")
		}

		k := rand.Intn(workerno - i)
		//k := workerno - i - 1
		// Playing it safe
		var branchcopy = make([]uint16, consts.DISTRIBUTIONROWS)
		copy(branchcopy, brancheslist[k])
		go worker(branchcopy, freethread, res)
		// Fill in the hole
		brancheslist[k] = brancheslist[workerno-i-1]
	}

	// Wait until last workers finished
	for j := 0; j < consts.PARALLELTHREADS; j++ {
		<- freethread
	}

}
